module advance_it_h_mod
 implicit none 
 real,private :: xscale,zscale
 real, pointer,private :: dvv(:)
 integer,private :: n1,n2,i1_src,i2_src
 real,private :: fmax,fpeak,dt,dx,dz
 logical, private :: do_bound
 real, private, pointer :: b_b(:),b_e(:),a_b(:),a_e(:)
 
 contains 
 subroutine init_advance(d_x,d_z,d_t,vel,i_1,i_2,f_max,f_peak,bound)
   real :: d_x,d_z,d_t
   real :: vel(:,:)
   integer :: i1,i2,ix,iy,i_1,i_2,i
   real :: f_max,f_peak
   logical ::bound
   
   dt=d_t
   dx=d_x
   dz=d_z
   i1_src=i_1
   i2_src=i_2
   fmax=f_max
   fpeak=f_peak
   do_bound=bound 
   n1=size(vel,1)
   n2=size(vel,2)
   allocate(dvv(n1*n2))
  i=0
   do i2=1,n2
     do i1=1,n1
       i=i+1
       dvv(i)=(vel(i1,i2)*vel(i1,i2))
     end do
   end do
   xscale=.25*dt*dt/dx/dx
   zscale=.25*dt*dt/dz/dz
   
   allocate(b_b(20),b_e(20))
   allocate(a_b(20),a_e(20))
   do i1=1,20
     b_b(i1)=exp(-(.3/20.*real(21-i1))**2)
     b_e(21-i1)=b_b(i1)
    ! write(0,*) "xx",i1,b_b(i1)
   end do
   a_b=1./b_b
   a_e=1./b_e
   write(0,*) "XXX",b_e
 end subroutine
 subroutine advance_it(it,pm,p,pp,s,do_for,t_1,t_2,b_1,b_2)
  logical :: do_for
  integer it
  real :: pm(:,:),p(:,:),pp(:,:),s(:,:),t_1(:,:),t_2(:,:),b_1(:,:),b_2(:,:)
   call advance_it_h(n1*n2,it,pm,p,pp,s)
end subroutine
 subroutine advance_it_h(n,it,pm,p,pp,s)
  integer ::n
  real :: pm(n),p(n),pp(n),s(n)
  integer :: i1,i2,it,i,j1,j2,j3,j4,j5
  logical :: do_for
  real  :: q,a1x,a1z,a2x,a2z,a3x,a3z,a4x,a4z,a5x,a5z,a0,x
  real z_d(11,5),x_d(11,5)
  integer offset(11,5),ic1,ic2
  real :: dtx,dtz
  real deriv1,deriv2;
  
  dtx=dt*dt/dx/dx
  dtz=dt*dt/dz/dz
  j1=n1
  j2=n1*2
  j3=n1*3
  j4=n1*4
  j5=n1*5
  
  a1x=3.333333333*dt*dt/(2*dx*dx);
        a1z=3.333333333*dt*dt/(2*dz*dz);

        a2x=-.4761904762*dt*dt/(2*dx*dx);
        a2z=-.4761904762*dt*dt/(2*dz*dz);

        a3x=.0793650794*dt*dt/(2*dx*dx);
        a3z=.0793650794*dt*dt/(2*dz*dz);

        a4x=-.0099206349*dt*dt/(2*dx*dx);
        a4z=-.0099206349*dt*dt/(2*dz*dz);

        a5x=.0006349206*dt*dt/(2*dx*dx);
        a5z=.0006349206*dt*dt/(2*dz*dz);

        a0=a5x+a5z+a4x+a4z+a3x+a3z+a2x+a2z+a1x+a1z;
        a0=-2.*a0;
        z_d=0; x_d=0; offset=0
        

  !$OMP PARALLEL DO default(shared) private(i1,x)
    do i1=j5+5,n1*n2-j5-5
       x=dvv(i1)*(&
             p(i1)*a0+&
             a1x*(p(i1+j1)+p(i1-j1))+&
             a2x*(p(i1+j2)+p(i1-j2))+&
             a3x*(p(i1+j3)+p(i1-j3))+&
             a4x*(p(i1+j4)+p(i1-j4))+&
             a5x*(p(i1+j5)+p(i1-j5))+&
             a1z*(p(i1+1)+p(i1-1))+&
             a2z*(p(i1+2)+p(i1-2))+&
             a3z*(p(i1+3)+p(i1-3))+&
             a4z*(p(i1+4)+p(i1-4))+&
             a5z*(p(i1+5)+p(i1-5)))
             pp(i1)=-pm(i1)+2.0*p(i1)+x+s(i1)

  end do
  !$OMP END PARALLEL DO


 end subroutine

subroutine bound_it(pm,p,pp,b,e)
  real :: pp(:,:),pm(:,:),p(:,:),b(:),e(:)
  integer :: i1,i2
  do i2=1,20
    do i1=1,n1
      pm(i1,i2)=pm(i1,i2)*b(i2)
       p(i1,i2)=p(i1,i2)*b(i2)
      pp(i1,i2)=pp(i1,i2)*b(i2)

    end do
  end do
  do i2=1,20
    do i1=1,n1
      pm(i1,i2+n2-20)=pm(i1,i2+n2-20)*e(i2)
      p(i1,i2+n2-20)=p(i1,i2+n2-20)*e(i2)
      pp(i1,i2+n2-20)=pp(i1,i2+n2-20)*e(i2)

    end do
  end do
  do i2=1,n2
    do i1=1,20
      pm(i1,i2)=pm(i1,i2)*b(i1)
       p(i1,i2)=p (i1,i2)*b(i1)
      pp(i1,i2)=pp(i1,i2)*b(i1)

    end do
  end do
  do i2=1,n2
    do i1=1,20
      pm(i1+n1-20,i2)=pm(i1+n1-20,i2)*e(i1)
      p (i1+n1-20,i2)=p (i1+n1-20,i2)*e(i1)
      pp(i1+n1-20,i2)=pp(i1+n1-20,i2)*e(i1)

    end do
  end do

end subroutine


subroutine form_source(s,t)
  real :: s(:,:)
 real :: t
  real :: tdelay
  real :: ts,xn,zn,x
  integer :: i1,i2
  
  s=0;
  tdelay=1./fpeak
  if(t<=2*tdelay) then
    x=atan(1.)*2*fpeak*(t-tdelay)
    ts=exp(-x*x)*(1.-2.*x*x)
    do i2=i2_src-3,i2_src+3
      xn=i2-i2_src
      do i1=i1_src-3,i1_src+3
        zn=i1-i1_src
        s(i1,i2)=ts*exp(-xn*xn-zn*zn)
      end do
    end do
  end if
  
    

end subroutine


end module
